@@ -17,6 +17,8 @@
 
 #include <inttypes.h>
 
+#include <isc/histo.h>
+
 #include <dns/types.h>
 
 /*%
@@ -157,310 +159,49 @@ enum {
 };
 
 /*%
- * Traffic size statistics counters. Used as isc_statscounter_t values.
+ * Traffic size statistics, according to RSSAC002 section 2.4
+ * https://www.icann.org/en/system/files/files/rssac-002-measurements-root-20nov14-en.pdf
+ *
+ * The RSSAC002 linear bucketing does not directly match the log-linear
+ * bucketing of an `isc_histo_t`, so we need to adjust some parameters
+ * to fit.
+ *
+ * To map a message size to an `isc_histo_t`, first divide by
+ * DNS_SIZEHISTO_QUANTUM so that `isc_histo_inc()` is presented with
+ * one value per RSSAC002 bucket.
+ *
+ * Configure the `isc_histo_t` with large enough `sigbits` that its
+ * one-value-per-bucket range (its `UNITBUCKETS`) covers the range
+ * required by RSSAC002.
+ */
+#define DNS_SIZEHISTO_QUANTUM	 16
+#define DNS_SIZEHISTO_MAXIN	 (288 / DNS_SIZEHISTO_QUANTUM)
+#define DNS_SIZEHISTO_MAXOUT	 (4096 / DNS_SIZEHISTO_QUANTUM)
+#define DNS_SIZEHISTO_SIGBITSIN	 4
+#define DNS_SIZEHISTO_SIGBITSOUT 7
+
+#define DNS_SIZEHISTO_BUCKETIN(size) \
+	ISC_MIN(size / DNS_SIZEHISTO_QUANTUM, DNS_SIZEHISTO_MAXIN)
+
+#define DNS_SIZEHISTO_BUCKETOUT(size) \
+	ISC_MIN(size / DNS_SIZEHISTO_QUANTUM, DNS_SIZEHISTO_MAXOUT)
+
+STATIC_ASSERT(DNS_SIZEHISTO_MAXIN <=
+		      ISC_HISTO_UNITBUCKETS(DNS_SIZEHISTO_SIGBITSIN),
+	      "must be enough histogram buckets for RSSAC002");
+
+STATIC_ASSERT(DNS_SIZEHISTO_MAXOUT <=
+		      ISC_HISTO_UNITBUCKETS(DNS_SIZEHISTO_SIGBITSOUT),
+	      "must be enough histogram buckets for RSSAC002");
+
+/*
+ * For consistency with other stats counters
  */
 enum {
-	dns_sizecounter_in_0 = 0,
-	dns_sizecounter_in_16 = 1,
-	dns_sizecounter_in_32 = 2,
-	dns_sizecounter_in_48 = 3,
-	dns_sizecounter_in_64 = 4,
-	dns_sizecounter_in_80 = 5,
-	dns_sizecounter_in_96 = 6,
-	dns_sizecounter_in_112 = 7,
-	dns_sizecounter_in_128 = 8,
-	dns_sizecounter_in_144 = 9,
-	dns_sizecounter_in_160 = 10,
-	dns_sizecounter_in_176 = 11,
-	dns_sizecounter_in_192 = 12,
-	dns_sizecounter_in_208 = 13,
-	dns_sizecounter_in_224 = 14,
-	dns_sizecounter_in_240 = 15,
-	dns_sizecounter_in_256 = 16,
-	dns_sizecounter_in_272 = 17,
-	dns_sizecounter_in_288 = 18,
-
-	dns_sizecounter_in_max = 19,
+	dns_sizecounter_in_max = DNS_SIZEHISTO_MAXIN + 1,
+	dns_sizecounter_out_max = DNS_SIZEHISTO_MAXOUT + 1,
 };
 
-enum {
-	dns_sizecounter_out_0 = 0,
-	dns_sizecounter_out_16 = 1,
-	dns_sizecounter_out_32 = 2,
-	dns_sizecounter_out_48 = 3,
-	dns_sizecounter_out_64 = 4,
-	dns_sizecounter_out_80 = 5,
-	dns_sizecounter_out_96 = 6,
-	dns_sizecounter_out_112 = 7,
-	dns_sizecounter_out_128 = 8,
-	dns_sizecounter_out_144 = 9,
-	dns_sizecounter_out_160 = 10,
-	dns_sizecounter_out_176 = 11,
-	dns_sizecounter_out_192 = 12,
-	dns_sizecounter_out_208 = 13,
-	dns_sizecounter_out_224 = 14,
-	dns_sizecounter_out_240 = 15,
-	dns_sizecounter_out_256 = 16,
-	dns_sizecounter_out_272 = 17,
-	dns_sizecounter_out_288 = 18,
-	dns_sizecounter_out_304 = 19,
-	dns_sizecounter_out_320 = 20,
-	dns_sizecounter_out_336 = 21,
-	dns_sizecounter_out_352 = 22,
-	dns_sizecounter_out_368 = 23,
-	dns_sizecounter_out_384 = 24,
-	dns_sizecounter_out_400 = 25,
-	dns_sizecounter_out_416 = 26,
-	dns_sizecounter_out_432 = 27,
-	dns_sizecounter_out_448 = 28,
-	dns_sizecounter_out_464 = 29,
-	dns_sizecounter_out_480 = 30,
-	dns_sizecounter_out_496 = 31,
-	dns_sizecounter_out_512 = 32,
-	dns_sizecounter_out_528 = 33,
-	dns_sizecounter_out_544 = 34,
-	dns_sizecounter_out_560 = 35,
-	dns_sizecounter_out_576 = 36,
-	dns_sizecounter_out_592 = 37,
-	dns_sizecounter_out_608 = 38,
-	dns_sizecounter_out_624 = 39,
-	dns_sizecounter_out_640 = 40,
-	dns_sizecounter_out_656 = 41,
-	dns_sizecounter_out_672 = 42,
-	dns_sizecounter_out_688 = 43,
-	dns_sizecounter_out_704 = 44,
-	dns_sizecounter_out_720 = 45,
-	dns_sizecounter_out_736 = 46,
-	dns_sizecounter_out_752 = 47,
-	dns_sizecounter_out_768 = 48,
-	dns_sizecounter_out_784 = 49,
-	dns_sizecounter_out_800 = 50,
-	dns_sizecounter_out_816 = 51,
-	dns_sizecounter_out_832 = 52,
-	dns_sizecounter_out_848 = 53,
-	dns_sizecounter_out_864 = 54,
-	dns_sizecounter_out_880 = 55,
-	dns_sizecounter_out_896 = 56,
-	dns_sizecounter_out_912 = 57,
-	dns_sizecounter_out_928 = 58,
-	dns_sizecounter_out_944 = 59,
-	dns_sizecounter_out_960 = 60,
-	dns_sizecounter_out_976 = 61,
-	dns_sizecounter_out_992 = 62,
-	dns_sizecounter_out_1008 = 63,
-	dns_sizecounter_out_1024 = 64,
-	dns_sizecounter_out_1040 = 65,
-	dns_sizecounter_out_1056 = 66,
-	dns_sizecounter_out_1072 = 67,
-	dns_sizecounter_out_1088 = 68,
-	dns_sizecounter_out_1104 = 69,
-	dns_sizecounter_out_1120 = 70,
-	dns_sizecounter_out_1136 = 71,
-	dns_sizecounter_out_1152 = 72,
-	dns_sizecounter_out_1168 = 73,
-	dns_sizecounter_out_1184 = 74,
-	dns_sizecounter_out_1200 = 75,
-	dns_sizecounter_out_1216 = 76,
-	dns_sizecounter_out_1232 = 77,
-	dns_sizecounter_out_1248 = 78,
-	dns_sizecounter_out_1264 = 79,
-	dns_sizecounter_out_1280 = 80,
-	dns_sizecounter_out_1296 = 81,
-	dns_sizecounter_out_1312 = 82,
-	dns_sizecounter_out_1328 = 83,
-	dns_sizecounter_out_1344 = 84,
-	dns_sizecounter_out_1360 = 85,
-	dns_sizecounter_out_1376 = 86,
-	dns_sizecounter_out_1392 = 87,
-	dns_sizecounter_out_1408 = 88,
-	dns_sizecounter_out_1424 = 89,
-	dns_sizecounter_out_1440 = 90,
-	dns_sizecounter_out_1456 = 91,
-	dns_sizecounter_out_1472 = 92,
-	dns_sizecounter_out_1488 = 93,
-	dns_sizecounter_out_1504 = 94,
-	dns_sizecounter_out_1520 = 95,
-	dns_sizecounter_out_1536 = 96,
-	dns_sizecounter_out_1552 = 97,
-	dns_sizecounter_out_1568 = 98,
-	dns_sizecounter_out_1584 = 99,
-	dns_sizecounter_out_1600 = 100,
-	dns_sizecounter_out_1616 = 101,
-	dns_sizecounter_out_1632 = 102,
-	dns_sizecounter_out_1648 = 103,
-	dns_sizecounter_out_1664 = 104,
-	dns_sizecounter_out_1680 = 105,
-	dns_sizecounter_out_1696 = 106,
-	dns_sizecounter_out_1712 = 107,
-	dns_sizecounter_out_1728 = 108,
-	dns_sizecounter_out_1744 = 109,
-	dns_sizecounter_out_1760 = 110,
-	dns_sizecounter_out_1776 = 111,
-	dns_sizecounter_out_1792 = 112,
-	dns_sizecounter_out_1808 = 113,
-	dns_sizecounter_out_1824 = 114,
-	dns_sizecounter_out_1840 = 115,
-	dns_sizecounter_out_1856 = 116,
-	dns_sizecounter_out_1872 = 117,
-	dns_sizecounter_out_1888 = 118,
-	dns_sizecounter_out_1904 = 119,
-	dns_sizecounter_out_1920 = 120,
-	dns_sizecounter_out_1936 = 121,
-	dns_sizecounter_out_1952 = 122,
-	dns_sizecounter_out_1968 = 123,
-	dns_sizecounter_out_1984 = 124,
-	dns_sizecounter_out_2000 = 125,
-	dns_sizecounter_out_2016 = 126,
-	dns_sizecounter_out_2032 = 127,
-	dns_sizecounter_out_2048 = 128,
-	dns_sizecounter_out_2064 = 129,
-	dns_sizecounter_out_2080 = 130,
-	dns_sizecounter_out_2096 = 131,
-	dns_sizecounter_out_2112 = 132,
-	dns_sizecounter_out_2128 = 133,
-	dns_sizecounter_out_2144 = 134,
-	dns_sizecounter_out_2160 = 135,
-	dns_sizecounter_out_2176 = 136,
-	dns_sizecounter_out_2192 = 137,
-	dns_sizecounter_out_2208 = 138,
-	dns_sizecounter_out_2224 = 139,
-	dns_sizecounter_out_2240 = 140,
-	dns_sizecounter_out_2256 = 141,
-	dns_sizecounter_out_2272 = 142,
-	dns_sizecounter_out_2288 = 143,
-	dns_sizecounter_out_2304 = 144,
-	dns_sizecounter_out_2320 = 145,
-	dns_sizecounter_out_2336 = 146,
-	dns_sizecounter_out_2352 = 147,
-	dns_sizecounter_out_2368 = 148,
-	dns_sizecounter_out_2384 = 149,
-	dns_sizecounter_out_2400 = 150,
-	dns_sizecounter_out_2416 = 151,
-	dns_sizecounter_out_2432 = 152,
-	dns_sizecounter_out_2448 = 153,
-	dns_sizecounter_out_2464 = 154,
-	dns_sizecounter_out_2480 = 155,
-	dns_sizecounter_out_2496 = 156,
-	dns_sizecounter_out_2512 = 157,
-	dns_sizecounter_out_2528 = 158,
-	dns_sizecounter_out_2544 = 159,
-	dns_sizecounter_out_2560 = 160,
-	dns_sizecounter_out_2576 = 161,
-	dns_sizecounter_out_2592 = 162,
-	dns_sizecounter_out_2608 = 163,
-	dns_sizecounter_out_2624 = 164,
-	dns_sizecounter_out_2640 = 165,
-	dns_sizecounter_out_2656 = 166,
-	dns_sizecounter_out_2672 = 167,
-	dns_sizecounter_out_2688 = 168,
-	dns_sizecounter_out_2704 = 169,
-	dns_sizecounter_out_2720 = 170,
-	dns_sizecounter_out_2736 = 171,
-	dns_sizecounter_out_2752 = 172,
-	dns_sizecounter_out_2768 = 173,
-	dns_sizecounter_out_2784 = 174,
-	dns_sizecounter_out_2800 = 175,
-	dns_sizecounter_out_2816 = 176,
-	dns_sizecounter_out_2832 = 177,
-	dns_sizecounter_out_2848 = 178,
-	dns_sizecounter_out_2864 = 179,
-	dns_sizecounter_out_2880 = 180,
-	dns_sizecounter_out_2896 = 181,
-	dns_sizecounter_out_2912 = 182,
-	dns_sizecounter_out_2928 = 183,
-	dns_sizecounter_out_2944 = 184,
-	dns_sizecounter_out_2960 = 185,
-	dns_sizecounter_out_2976 = 186,
-	dns_sizecounter_out_2992 = 187,
-	dns_sizecounter_out_3008 = 188,
-	dns_sizecounter_out_3024 = 189,
-	dns_sizecounter_out_3040 = 190,
-	dns_sizecounter_out_3056 = 191,
-	dns_sizecounter_out_3072 = 192,
-	dns_sizecounter_out_3088 = 193,
-	dns_sizecounter_out_3104 = 194,
-	dns_sizecounter_out_3120 = 195,
-	dns_sizecounter_out_3136 = 196,
-	dns_sizecounter_out_3152 = 197,
-	dns_sizecounter_out_3168 = 198,
-	dns_sizecounter_out_3184 = 199,
-	dns_sizecounter_out_3200 = 200,
-	dns_sizecounter_out_3216 = 201,
-	dns_sizecounter_out_3232 = 202,
-	dns_sizecounter_out_3248 = 203,
-	dns_sizecounter_out_3264 = 204,
-	dns_sizecounter_out_3280 = 205,
-	dns_sizecounter_out_3296 = 206,
-	dns_sizecounter_out_3312 = 207,
-	dns_sizecounter_out_3328 = 208,
-	dns_sizecounter_out_3344 = 209,
-	dns_sizecounter_out_3360 = 210,
-	dns_sizecounter_out_3376 = 211,
-	dns_sizecounter_out_3392 = 212,
-	dns_sizecounter_out_3408 = 213,
-	dns_sizecounter_out_3424 = 214,
-	dns_sizecounter_out_3440 = 215,
-	dns_sizecounter_out_3456 = 216,
-	dns_sizecounter_out_3472 = 217,
-	dns_sizecounter_out_3488 = 218,
-	dns_sizecounter_out_3504 = 219,
-	dns_sizecounter_out_3520 = 220,
-	dns_sizecounter_out_3536 = 221,
-	dns_sizecounter_out_3552 = 222,
-	dns_sizecounter_out_3568 = 223,
-	dns_sizecounter_out_3584 = 224,
-	dns_sizecounter_out_3600 = 225,
-	dns_sizecounter_out_3616 = 226,
-	dns_sizecounter_out_3632 = 227,
-	dns_sizecounter_out_3648 = 228,
-	dns_sizecounter_out_3664 = 229,
-	dns_sizecounter_out_3680 = 230,
-	dns_sizecounter_out_3696 = 231,
-	dns_sizecounter_out_3712 = 232,
-	dns_sizecounter_out_3728 = 233,
-	dns_sizecounter_out_3744 = 234,
-	dns_sizecounter_out_3760 = 235,
-	dns_sizecounter_out_3776 = 236,
-	dns_sizecounter_out_3792 = 237,
-	dns_sizecounter_out_3808 = 238,
-	dns_sizecounter_out_3824 = 239,
-	dns_sizecounter_out_3840 = 240,
-	dns_sizecounter_out_3856 = 241,
-	dns_sizecounter_out_3872 = 242,
-	dns_sizecounter_out_3888 = 243,
-	dns_sizecounter_out_3904 = 244,
-	dns_sizecounter_out_3920 = 245,
-	dns_sizecounter_out_3936 = 246,
-	dns_sizecounter_out_3952 = 247,
-	dns_sizecounter_out_3968 = 248,
-	dns_sizecounter_out_3984 = 249,
-	dns_sizecounter_out_4000 = 250,
-	dns_sizecounter_out_4016 = 251,
-	dns_sizecounter_out_4032 = 252,
-	dns_sizecounter_out_4048 = 253,
-	dns_sizecounter_out_4064 = 254,
-	dns_sizecounter_out_4080 = 255,
-	dns_sizecounter_out_4096 = 256,
-
-	dns_sizecounter_out_max = 257
-};
-
-#define DNS_STATS_NCOUNTERS 8
-
-#if 0
-/*%<
- * Flag(s) for dns_xxxstats_dump().  DNS_STATSDUMP_VERBOSE is obsolete.
- * ISC_STATSDUMP_VERBOSE should be used instead.  These two values are
- * intentionally defined to be the same value to ensure binary compatibility.
- */
-#define DNS_STATSDUMP_VERBOSE 0x00000001 /*%< dump 0-value counters */
-#endif					 /* if 0 */
-
-/*%<
- * (Obsoleted)
- */
-extern const char *dns_statscounter_names[];
-
 /*%
  * Attributes for statistics counters of RRset and Rdatatype types.
  *
@@ -493,7 +234,7 @@ extern const char *dns_statscounter_names[];
 /*%<
  * Conversion macros among dns_rdatatype_t, attributes and isc_statscounter_t.
  */
-#define DNS_RDATASTATSTYPE_BASE(type)  ((dns_rdatatype_t)((type)&0xFFFF))
+#define DNS_RDATASTATSTYPE_BASE(type)  ((dns_rdatatype_t)((type) & 0xFFFF))
 #define DNS_RDATASTATSTYPE_ATTR(type)  ((type) >> 16)
 #define DNS_RDATASTATSTYPE_VALUE(b, a) (((a) << 16) | (b))
 
@@ -511,13 +252,13 @@ typedef enum {
 typedef void (*dns_generalstats_dumper_t)(isc_statscounter_t, uint64_t, void *);
 typedef void (*dns_rdatatypestats_dumper_t)(dns_rdatastatstype_t, uint64_t,
 					    void *);
-typedef void (*dns_dnssecsignstats_dumper_t)(dns_keytag_t, uint64_t, void *);
+typedef void (*dns_dnssecsignstats_dumper_t)(uint32_t, uint64_t, void *);
 typedef void (*dns_opcodestats_dumper_t)(dns_opcode_t, uint64_t, void *);
 typedef void (*dns_rcodestats_dumper_t)(dns_rcode_t, uint64_t, void *);
 
 ISC_LANG_BEGINDECLS
 
-isc_result_t
+void
 dns_generalstats_create(isc_mem_t *mctx, dns_stats_t **statsp, int ncounters);
 /*%<
  * Create a statistics counter structure of general type.  It counts a general
@@ -529,14 +270,9 @@ dns_generalstats_create(isc_mem_t *mctx, dns_stats_t **statsp, int ncounters);
  *\li	'mctx' must be a valid memory context.
  *
  *\li	'statsp' != NULL && '*statsp' == NULL.
- *
- * Returns:
- *\li	ISC_R_SUCCESS	-- all ok
- *
- *\li	anything else	-- failure
  */
 
-isc_result_t
+void
 dns_rdatatypestats_create(isc_mem_t *mctx, dns_stats_t **statsp);
 /*%<
  * Create a statistics counter structure per rdatatype.
@@ -545,14 +281,9 @@ dns_rdatatypestats_create(isc_mem_t *mctx, dns_stats_t **statsp);
  *\li	'mctx' must be a valid memory context.
  *
  *\li	'statsp' != NULL && '*statsp' == NULL.
- *
- * Returns:
- *\li	ISC_R_SUCCESS	-- all ok
- *
- *\li	anything else	-- failure
  */
 
-isc_result_t
+void
 dns_rdatasetstats_create(isc_mem_t *mctx, dns_stats_t **statsp);
 /*%<
  * Create a statistics counter structure per RRset.
@@ -561,14 +292,9 @@ dns_rdatasetstats_create(isc_mem_t *mctx, dns_stats_t **statsp);
  *\li	'mctx' must be a valid memory context.
  *
  *\li	'statsp' != NULL && '*statsp' == NULL.
- *
- * Returns:
- *\li	ISC_R_SUCCESS	-- all ok
- *
- *\li	anything else	-- failure
  */
 
-isc_result_t
+void
 dns_opcodestats_create(isc_mem_t *mctx, dns_stats_t **statsp);
 /*%<
  * Create a statistics counter structure per opcode.
@@ -577,14 +303,9 @@ dns_opcodestats_create(isc_mem_t *mctx, dns_stats_t **statsp);
  *\li	'mctx' must be a valid memory context.
  *
  *\li	'statsp' != NULL && '*statsp' == NULL.
- *
- * Returns:
- *\li	ISC_R_SUCCESS	-- all ok
- *
- *\li	anything else	-- failure
  */
 
-isc_result_t
+void
 dns_rcodestats_create(isc_mem_t *mctx, dns_stats_t **statsp);
 /*%<
  * Create a statistics counter structure per assigned rcode.
@@ -593,14 +314,9 @@ dns_rcodestats_create(isc_mem_t *mctx, dns_stats_t **statsp);
  *\li	'mctx' must be a valid memory context.
  *
  *\li	'statsp' != NULL && '*statsp' == NULL.
- *
- * Returns:
- *\li	ISC_R_SUCCESS	-- all ok
- *
- *\li	anything else	-- failure
  */
 
-isc_result_t
+void
 dns_dnssecsignstats_create(isc_mem_t *mctx, dns_stats_t **statsp);
 /*%<
  * Create a statistics counter structure per assigned DNSKEY id.
@@ -609,11 +325,6 @@ dns_dnssecsignstats_create(isc_mem_t *mctx, dns_stats_t **statsp);
  *\li	'mctx' must be a valid memory context.
  *
  *\li	'statsp' != NULL && '*statsp' == NULL.
- *
- * Returns:
- *\li	ISC_R_SUCCESS	-- all ok
- *
- *\li	anything else	-- failure
  */
 
 void
@@ -804,22 +515,4 @@ dns_rcodestats_dump(dns_stats_t *stats, dns_rcodestats_dumper_t dump_fn,
  *\li	'stats' is a valid dns_stats_t created by dns_generalstats_create().
  */
 
-isc_result_t
-dns_stats_alloccounters(isc_mem_t *mctx, uint64_t **ctrp);
-/*%<
- * Allocate an array of query statistics counters from the memory
- * context 'mctx'.
- *
- * This function is obsoleted.  Use dns_xxxstats_create() instead.
- */
-
-void
-dns_stats_freecounters(isc_mem_t *mctx, uint64_t **ctrp);
-/*%<
- * Free an array of query statistics counters allocated from the memory
- * context 'mctx'.
- *
- * This function is obsoleted.  Use dns_stats_destroy() instead.
- */
-
 ISC_LANG_ENDDECLS
